	.opt proc6502,caseinsensitive,swapbin
;* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;*                                                                             *
;*                       C-128 80 COLUMN DISPLAY MANAGER                       *
;*                                                                             *
;* This program implements functions to efficiently manage multiple screens in *
;* the C-128's 80 column display system.  Full functionality is available on a *
;* C-128 with 64KB of video RAM.  A machine with 16KB of video RAM is limited  *
;* to buffering one screen & display blanking.  Automatic video RAM sizing is  *
;* performed during full initialization.                                       *
;*                                                                             *
;*  *
;*                                                                             *
;*      COPYRIGHT 1987 by BCS TECHNOLOGY LIMITED.   ALL RIGHTS RESERVED.      *
;*                                                                             *
;* Permission is hereby granted to use, copy, modify and distribute this soft- *
;* ware, provided this copyright notice remains unaltered in the source code   *
;* and proper attribution is given.  Redistribution in any form must be at no  *
;* charge to the end user.  This code or any part thereof, including any deri- *
;* vation, MAY NOT be incorporated into any package intended for sale unless   *
;* written permission to do so has been granted by the copyright holder.       *
;*                                                                             *
;*  *
;*                                                                             *
;*        !!! THERE IS NO WARRANTY OF ANY KIND WITH THIS SOFTWARE !!!          *
;*                                                                             *
;*  While it is believed that all code will perform as intended, the user as-  *
;*  sumes all risk in connection with the incorporation of this software into  *
;*  any system.  If any provision set forth herein is not acceptable to you,   *
;*  do not use this software and immediately delete it from your system.       *
;*                                                                             *
;* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;
;* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;*                                                                             *
;* C-128 DISPLAY MANAGER FUNCTIONS                                             *
;* ===============================                                             *
;*                                                                             *
;*                                                                        64K  *
;* Func  Action Performed                                          VRD   Req'd *
;*  *
;*   0   Initialization, required at startup                       N/A    N/A  *
;*   1   Copy screen to screen                                     Yes     No  *
;*   2   Exchange active/inactive screens                          Yes    Yes  *
;*   3   Redirect display                                           No     No  *
;*   4   Enable/disable status line                                Yes    Yes  *
;*   5   Enable/disable visible display                            Yes     No  *
;*   6   Adjust vertical raster position                           Yes    Yes  *
;*   7   Copy screen to RAM-0                                       No     No  *
;*   8   Reserved function                                         N/A    N/A  *
;*   9   Copy RAM-0 to screen                                      Yes     No  *
;*  *
;*                                                                             *
;* Functions with YES in VRD (vertical retrace delay) column can be made to    *
;* wait until the VDC is in the vertical retrace blanking part of the raster   *
;* generation process by ORing the function number with $80.  Functions that   *
;* cannot recognize the vertical retrace period will ignore such a request.    *
;*                                                                             *
;* Separate entry calls are used to write text to the status line & test the   *
;* status of the semaphore.  See also the presence signature pointers.  Refer  *
;* to the documentation PDF included with this source code for more details.   *
;*                                                                             *
;* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;
;* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;*                                                                             *
;*                      DISPLAY MANAGER ACCESS JUMP TABLE                      *
;*                =============================================                *
;*                JDPYMGR    $0E000  Numbered Manager Functions                *
;*                DMSIGPTR   $0E003  Signature String Pointer                  *
;*                DMVERSON   $0E005  Version & Copyright Notice                *
;*                JDSLINE    $0E007  Status Line Text Writer                   *
;*                JTSTSEM    $0E00A  Semaphore Status Checker                  *
;*                =============================================                *
;*                                                                             *
;* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;
;===============================================================================
;
;	        * * * * * * * * * * * * *
;	        * Comment Abbreviations *
;	        * * * * * * * * * * * * *
;	
;	ARAM   attribute video RAM
;	 CBM   Commodore Business Machines
;	DRAM   display video RAM
;	KRAM   character (font) video RAM
;	 LSB   least significant byte
;	 MMU   8722 memory management unit
;	 MPU   8502 microprocessor
;	 MSB   most significant byte
;	 VDC   8563/8568 video display controller
;	VRAM   video RAM
;	 VRT   vertical retrace blanking period
;	  ZP   zero page
;	
;
;	    * * * * * * * * *
;	    * MPU Registers *
;	    * * * * * * * * *
;	
;	  .A   accumulator
;	  .c   carry status
;	  .n   result negative
;	  SP   stack pointer
;	  SR   status
;	  .X   X-index
;	  .Y   Y-index
;	  .z   result zero
;	
;
;===============================================================================
;
;	* * * * * * * * * * *
;	* REVISION  HISTORY *                                         
;	* * * * * * * * * * *
;
;Ver    Rev Date    Description
;
;1.1   2008/03/17   Pulled out of mothballs to be published on-line.  Added the
;                   the following features:
;
;                   (a) Copy a screen to/from system RAM, making it possible to
;                       save a screen to disk for later recall, a feature that
;                       works very well with the Xetec Lt. Kernal hard disk sub-
;                       system or with an REU.  :)
;
;                   (b) Vertical retrace delay option, which eliminates display
;                       tearing when switching screens.  I made this an option
;                       because it does have an effect on performance & someone
;                       might be more concerned with that than with visual mat-
;                       ters.  The original code had this feature permanently
;                       enabled.
;
;                   Internal changes made to improve operation include:
;
;                   (a) Removed automatic display repositioning in the status
;                       line enable/disable function.  A manual function is
;                       available for that purpose.
;
;                   (b) Removed limits from raster adjustment features.  If a
;                       user is so inclined, he can adjust the picture right
;                       off the screen.  :)
;
;                   (c) Rearranged access table to place signature & copyright
;                       strings before fixed access entries.
;
;1.0   1987/11/09   Original version, which was part of a vertical application
;                   package for a truck leasing & billing system.
;
;
;===============================================================================
;
;C-128 HARDWARE ASSIGNMENTS
;
mmu      =$d500                ;8722 memory management unit (I/O)
mmuhi    =$ff00                ;8722 memory management unit (hi RAM)
vdc      =$d600                ;8563/68 video display controller
;
;===============================================================================
;
;8563/68 VIDEO DISPLAY CONTROLLER DECLARATIONS
;
vdcadr   =vdc                  ;setup (write)/status (read) port...
;
;			
;			write: selects internal register to access
;			read:  reports VDC status...
;
;			xxx11xxx
;			|||  |||
;			|||  +++> chip version (meaning unknown)
;			||+> 1 = vertical retrace in progress
;			|+> 1 = light pen location latched
;			+> 1 = selected register ready for access
;
;			INTR is asserted in the 8568 when bit 7 is set & is
;			deasserted when the status register is read.  INTR
;			is not connected in the C-128DCR.
;			
; 
vdcdat   =vdcadr+1             ;read/write register port
;
;                                       
;
;	
;	| 8563/68 VDC Internal Registers |
;	
;
coltot   =0                    ;total character columns
colvis   =1                    ;visible character columns
hsypos   =2                    ;horizontal sync position
hvsyw    =3                    ;horizontal & vertical sync width...
;
;			xxxxxxxx
;			||||||||
;			||||++++> horizontal sync pulse
;			++++> vertical sync pulse
;
rowtot   =4                    ;total character rows
vfadj    =5                    ;vertical fine adjustment
rowvis   =6                    ;visible character rows
vsypos   =7                    ;vertical sync position
ilcctl   =8                    ;interlace mode control...
;
;			11111100
;			      ||
;			      ++> 00 = non-interlace (default)
;			             01 = interlace sync (double offset scan)
;			             10 = no interlace
;			             11 = interlace sync & video (alternating half-frames)
;
nsclpc   =9                    ;scan lines per character
csrctl   =10                   ;cursor mode control
csrbsl   =11                   ;cursor bottom scan line
dmemhi   =12                   ;start of SRAM (MSB)
dmemlo   =13                   ;start of SRAM (LSB)
cposhi   =14                   ;cursor position (MSB)
cposlo   =15                   ;cursor position (LSB)
lpvpos   =16                   ;light pen vertical position
lphpos   =17                   ;light pen horizontal position
vramhi   =18                   ;VRAM address (MSB), autoincrements
vramlo   =19                   ;VRAM address (LSB), autoincrements
amemhi   =20                   ;start of ARAM (MSB)
amemlo   =21                   ;start of ARAM (LSB)
hcsiz    =22                   ;character horizontal size...
;
;			xxxxxxxx
;			||||||||
;			||||++++> pixels per display character: default = %00001000
;			++++> pixels in character cell:     default = %01110000
;
vcsiz    =23                   ;character vertical size control...
;
;			111xxxxx
;			   |||||
;			   +++++> scan lines in character cell: default = %00001000
;
vsscr    =24                   ;vertical smooth scroll & control...
;
;			xxxxxxxx
;			||||||||
;			|||+++++> vertical smooth scroll
;			||+> 1 = slow character blinking rate (default)
;			|+> 1 = foreground/background reversed (default = 0)
;			+> 0 = fill VRAM, 1 = copy VRAM
;
hsscr    =25                   ;horizontal smooth scroll & control...
;
;			xxxxxxxx
;			||||||||
;			||||++++> horizontal smooth scroll
;			|||+> 1 = double pixel size
;			||+> 1 = semi-graphic mode on
;			|+> 1 = character attribute mode on (default)
;			+> 0 = character mode (default), 1 = bitmap mode
;
fgbgat   =26                   ;foreground/background color...
;
;			xxxxxxxx
;			||||||||
;			||||++++> background color (RGBI)
;			++++> foreground color (RGBI)
;
rowain   =27                   ;row address increment value
kramad   =28                   ;KRAM start address & VRAM type...
;
;			xxxx1111
;			||||
;			|||+> 0 = 4416 DRAM, 1 = 4464
;			+++> KRAM page address 
;
ulslp    =29                   ;underline scan line position
blkbyt   =30                   ;number of bytes for block copy/fill
vramrw   =31                   ;VRAM read/write access
bsrchi   =32                   ;block copy source address (MSB)
bsrclo   =33                   ;block copy source address (LSB)
hbposl   =34                   ;left horizontal blanking position
hbposr   =35                   ;right horizontal blanking position
dramrf   =36                   ;VRAM refresh cycles per scan line
;
;===============================================================================
;
;8722 MEMORY MANAGEMENT UNIT DECLARATIONS
;
mmucr    =mmuhi                ;configuration...
;
;			0xxxxxxx
;			 |||||||
;			 ||||||+> $D000-$DFFF:  0 = I/O hardware
;			 ||||||                    1 = RAM or character ROM
;			 |||||+> $4000-$7FFF:  0 = system ROM
;			 |||||                     1 = RAM
;			 |||++> $8000-$BFFF: 00 = system ROM
;			 |||                      01 = internal function ROM
;			 |||                      10 = external function ROM
;			 |||                      11 = RAM
;			 |++> $C000-$FFFF: 00 = system ROM
;			 |                        01 = internal function ROM
;			 |                        10 = external function ROM
;			 |                        11 = RAM
;			 +> RAM bank:     0 = RAM-0
;			                           1 = RAM-1
;
pcra     =mmu+1                ;preconfiguration A
pcrb     =mmu+2                ;preconfiguration B
mmurcr   =mmu+6                ;memory map configuration...
;
;			0x00xxxx
;			 |  ||||
;			 |  ||++> common RAM size: 00 =  1K
;			 |  ||                        01 =  4K
;			 |  ||                        10 =  8K
;			 |  ||                        11 = 16K
;			 |  ++> common RAM area: 00 = none
;			 |                            01 = bottom
;			 |                            10 = top
;			 |                            11 = both
;			 +> VIC video bank: 0 = RAM-0
;			                             1 = RAM-1
;
lcra     =mmucr+1              ;select configuration A
lcrb     =mmucr+2              ;select configuration B
lcrc     =mmucr+3              ;select configuration C
lcrd     =mmucr+4              ;select configuration D
;
;===============================================================================
;
;C-128 80 COLUMN DISPLAY MANAGER DECLARATIONS
;
_origin_ =$0e000               ;program load address (RAM-0)
;
;	-------------
;	Data Type Size Declarations
;	-------------
;
s_byte   =1                    ;byte or char
s_bits   =8                    ;bits in a byte
s_word   =2                    ;word
s_xword  =3                    ;extended word
s_page   =256                  ;65xx RAM page
;
;	----------
;	Local ZP Assignments
;	----------
;
zpptra   =$70                  ;pointer A
zpptrb   =$72                  ;pointer B
zpptrc   =$74                  ;pointer C
zpptrd   =$76                  ;pointer D
zpptre   =$78                  ;pointer E
zpptrf   =$7a                  ;pointer F
;
;	----------------
;	Arithmetic & Temporary Workspace
;	----------------
;
fac1     =$63                  ;floating accumulator #1
fac2     =$6a                  ;floating accumulator #2
imparm   =$ce                  ;temporary text pointer
;
;	-------------------
;	Screen Kernel Active Display Locations
;	-------------------
;
mode     =$d7                  ;active display...
;
;			x0000000
;			|
;			+> 1 = 80 column active
;
asvtbeg  =$e0                  ;active screen ZP variables begin
asvtend  =$f9                  ;active screen ZP variables end
colinpts =$e9                  ;start of input cursor column
curmod   =$0a2b                ;VDC cursor mode shadow
curypos  =$eb                  ;absolute cursor row
curxpos  =$ec                  ;absolute cursor column
lnkmap   =$035e                ;logical line link bitmap
lnkmape  =$0361                ;end of logical line link bitmap
rowaadr  =$e2                  ;absolute row attribute address
rowdadr  =$e0                  ;absolute row display address
rowinpts =$e8                  ;start of input cursor row
scbot    =$e4                  ;window bottom margin
sctop    =$e5                  ;window top margin
scleft   =$e6                  ;window left margin
scright  =$e7                  ;window right margin
tabmap   =$0354                ;tab stop bitmap
tabmape  =$035d                ;end of tab stop bitmap
vm3      =$0a2e                ;VDC display RAM start page
vm4      =$0a2f                ;VDC attribute RAM start page
;
;	---------------------
;	| DISPLAY MANAGER VIDEO RAM ORGANIZATION |
;	-----------------------------------
;	The display manager organizes 64K VRAM machines as follows:
;
;	* Screens are numbered 0 through 11 inclusive, with 0 being the visible
;	  screen, 1-10 being user-saved screens & 11 being the exchange screen.
;
;	* Screens are saved in 4K-sized frames.
;
;	* Frames are split into 2 segments of 2K each, with the  lower  segment
;	  storing the screen codes & the upper segment storing the attributes.
;
;	* Frames start on S_BUFFRM boundaries, e.g., $4000, $5000, $6000, etc.
;
;	* Each segment starts on a 2K page boundary, e.g., $4000, $4800, $5000,
;	  $5800, etc.
;
;	* Live display attributes are moved to $1000 to free up VRAM for use
;	  by the status line function.  This change is required because the
;	  C-128s screen kernal only records the page boundary at which attri-
;	  bute RAM starts.
;
;	16K VRAM systems have only 1 frame available at $1000 & live attributes
;	are not relocated.  No status line is possible in a 16K VRAM system.
;	If you have a flat C-128 or the European C-128D model (not the same
;	as the North American C-128DCR), you really should consider upgrading
;	video RAM to the full 64KB.
;	-----------------------------------
;
frevr16s =$01000               ;free VRAM start (16K)
frevr64s =$04000               ;free VRAM start (64K)
frevr16e =$02000               ;free VRAM end +1 (16K)
frevr64e =$10000               ;free VRAM end +1 (64K)
bufpag16 =>frevr16s            ;base buffer page (16K)
bufpag64 =>frevr64s            ;base buffer page (64K)
altarpag =>frevr16s            ;alternate attribute page (64K)
s_buffrm =$1000                ;buffer frame size
s_bufseg =s_buffrm/2           ;buffer segment size
s_vr16   =frevr16e-frevr16s    ;total buffer space (16K)
s_vr64   =frevr64e-frevr64s    ;total buffer space (64K)
n_fram16 =s_vr16/s_buffrm      ;total buffer frames (16K)
n_fram64 =s_vr64/s_buffrm      ;total buffer frames (64K)
n_scrn16 =n_fram16             ;user-accessible screens (16K)
n_scrn64 =n_fram64-2           ;user-accessible screens (64K)
exchscrn =n_scrn64+1           ;exchange screen number
holdscrn =n_fram64             ;hold screen number
livescrn =0                    ;visible screen number
;
;	-----------------
;	VDC Register Setup & Control Values
;	-----------------
;
dpyblnkl =$56                  ;left margin blanking value
dpyblnkr =$06                  ;right margin blanking value
vctradji =1                    ;vertical adjustment increment
;
;	-----------
;	VRAM Sizing Test Values
;	-----------
;
ramsmall =16                   ;returned for 16K VRAM
ramlarge =64                   ;returned for 64K VRAM
testadra =frevr64s+s_page*2    ;1st VRAM test address
testadrb =testadra+s_page      ;2nd VRAM test address
testpata =%01011010            ;VRAM test pattern #1
testpatb =%10100101            ;VRAM test pattern #2
;
;	-----------
;	Size-of Declarations
;	-----------
;
s_funtab =funtabe-funtab       ;function dispatch table
s_lnktab =lnkmape-tabmap+1     ;line link & tab bitmaps
s_sfbtab =s_byte*exchscrn+1    ;saved fore/background color table
s_szptab =asvtend-asvtbeg+1    ;active screen ZP table
s_tabssf =s_byte*exchscrn+1    ;saved screen flag table
s_tabsvc =s_byte*exchscrn+1    ;saved screen shadow table
s_tabsvl =s_lnktab*exchscrn+1  ;saved screen line link table
s_tabszp =s_szptab*exchscrn+1  ;saved screen ZP table
s_vdcstg =vdcregse-vdcregs     ;VDC register storage
s_zpptrx =zpptrf-zpptra+s_word ;ZP protective storage
;
;	---------
;	ASCII Control Codes
;	---------
;
a_cr     =$0d                  ;<CR>  carriage return
a_nul    =$00                  ;<NUL> null
;
;	--------------
;	PETSCII Conversion Constants
;	--------------
;
petcodcl =%00011111            ;low PETSCII control range end
petcodch =%10011111            ;high PETSCII control range end
petcodl  =%01100000            ;regular LC PETSCII range end
petcodm  =%10000000            ;alternate LC PETSCII range end
petcodp  =%01111111            ;pi PETSCII test value
petcodps =%01011110            ;pi screen code
scblank  =%00100000            ;blank char screen code
;
;	--------
;	Error Exit Codes
;	--------
;
e_busy   =$00                  ;busy
e_badsrc =$01                  ;bad source
e_badtgt =$02                  ;bad target
e_srctgt =$03                  ;target == source
e_badcs  =$04                  ;bad switch
e_range  =$05                  ;parm range/string too long
e_mode   =$40                  ;bot 80 columns
e_badop  =$80                  ;illegal function
e_init   =$c0                  ;not initialized
;
;	----------
;	Control & Test Masks
;	----------
;
krdbmask =%11110000            ;VDC KRAM reg dead bits
pschmask =%01111111            ;PETSCII to screen code high-range
psclmask =%10111111            ;PETSCII to screen code mid-range
pscmmask =%11011111            ;PETSCII to screen code low-range
semamask =%10000000            ;semaphore set
slatmask =%01110000            ;status line attribute test
slofmask =%11111110            ;status line off
slonmask =%00000001            ;status line on
slrvmask =%01000000            ;VDC reverse attribute
sw64mask =%01000000            ;status word: 64K VRAM flag
swfimask =%10000000            ;status word: full init flag
testvrt  =%00010000            ;VDC DRAM type
vbfsmask =%10000000            ;valid saved screen data
vicbmask =%11101111            ;VIC screen blank enable
viccmask =%00000011            ;VIC clock & test enable
vrbcmask =%10000000            ;VDC block copy control
vrbfmask =%01111111            ;VDC block fill control
vrtmask  =%00100000            ;VDC vertical retrace
vrtemask =%01111111            ;function VRT enable exclude
vrtfmask =%10000000            ;function VRT enable check
vsbfmask =%01000000            ;valid saved screen buffer
vssfmask =vbfsmask|vsbfmask    ;valid saved screen
;
;	-------------
;	Memory Configuration Masks
;	-------------
;
r0iomask =%00111110            ;RAM-0 & I/O, default map
r0mask   =%00111111            ;RAM-0
romhmask =%00001110            ;RAM-0 & high ROM
;
;	-----------
;	Miscellaneous Constants
;	-----------
;
bitabs   =$2c                  ;BIT absolute opcode
dlchr    =$0c027               ;copy char ROM to VRAM
clscrn   =$0c142               ;screen kernel: clear display
func_all =1                    ;all functions available
func_min =0                    ;minimal functionality available
funcrang =s_funtab/s_word      ;allowed function number range
hwstack  =$00100               ;MPU hardware stack
maxmem0  =$01212               ;top of BASIC text +1 pointer
switchhi =1                    ;hi value for switched functions
switchlo =0                    ;lo value for switched functions
vcadjswd =$80                  ;vertical adjustment down switch
;
;===============================================================================
;
;C-128 80 COLUMN DISPLAY MANAGER ACCESS TABLE
;
         *=_origin_
;
jdpymgr  jmp dpymgr            ;numbered functions
;
dmsigptr .word signatr         ;presence signature pointer
dmverson .word copyrit         ;copyright notice pointer
;
jdsline  jmp dsline            ;status line writer
jtstsem  jmp tstsem            ;test semaphore
;
;===============================================================================
;
;enter: PROGRAM ENTRY HANDLER
;
;	
;	.A = function code or parameter
;	.X = parameter #1
;	.Y = parameter #2
;	.C = 0: called by a numbered function
;	     1: not called by a numbered function
;	
;	.A = entry value
;	.X = entry value
;	.Y = entry value
;	.C = 0
;	.N = 0
;	.Z = 0
;	
;	An immediate exit to the error handler will occur if the manager  has
;	not been initialized, the system is running on the 40 column side, or
;	if the semaphore has been raised due to another manager function hav-
;	ing been called.  The latter should prevent a conflict between an in-
;	process foreground call & an IRQ handler.  The semaphore check is by-
;	passed during initialization.
;	
;
enter    ora #0                ;is this an init operation?
         beq .0000010          ;yes, skip semaphore check
;
         bit semaphor          ;test semaphore state
         bmi .0000050          ;manager busy
;
;	
;	Set up system state
;	
;         
.0000010 php                   ;save carry state
         pha                   ;protect entry value
         lda #semamask
         sta semaphor          ;manager is busy
         lda mmucr             ;current memory map
         pha
         lda #r0iomask
         sta mmucr             ;select RAM-0 & I/O only
         pla                   ;is our entry memory map
         sta mmucrx            ;save old memory map
         lda #0
         sta vrtflag           ;reset VRD flag
         pla                   ;recover entry value
         sta reg_ats           ;save entry parameters
         stx reg_xts
         sty reg_yts
         tsx                   ;current stack pointer
	.rept 3
         inx
	.endr
         stx reg_sptr          ;now entry stack pointer
         pha                   ;save raw function code
         and #vrtfmask         ;extract VRT flag &...
         sta vrtflag           ;save, if used
         pla
         and #vrtemask         ;get rid of VRT flag
         sta funccode          ;now a function code
;
;	
;	Protect ZP locations we use
;	
;
         lda imparm
         sta imparmx
         lda imparm+s_byte
         sta imparmx+s_byte
         ldx #s_zpptrx-1
;
.0000020 lda zpptra,x
         sta zpptrx,x
         dex
         bpl .0000020
;
;	
;	Continue with entry procedure
;	
;
         plp                   ;recover carry
         bcs .0000030          ;not a function
;
         ldx funccode          ;function code
         beq .0000040          ;initializing
;
.0000030 bit dmstatus          ;check our init state
         bpl badinit           ;not initialized, so die...
;
;			-----------------
;			80CDM must be initialized before a
;			non-zero function can be invoked.
;			-----------------
;
.0000040 bit mode              ;runing on 40 or 80 columns?
         bpl badmode           ;40 columns  error
;
;	
;	Ready to process
;	
;
         lda #0
         pha
         txa                   ;get function code
         ldx reg_xts           ;recover...
         ldy reg_yts           ;entry parameters
         plp                   ;clear SR
         rts                   ;process function
;
;	
;	Handle a processing conflict
;	
;
.0000050 pla                   ;rebalance the...
         pla                   ;stack
         lda #e_busy           ;set error code
         sec
         rts                   ;back to caller
;
;===============================================================================
;
;ERROR EXIT HANDLER
;
badfunc  lda #e_badop          ;bad function
	.byte bitabs
badsrc   lda #e_badsrc         ;bad source screen
	.byte bitabs
badtgt   lda #e_badtgt         ;bad target screen
	.byte bitabs
badseqt  lda #e_srctgt         ;target = source
	.byte bitabs
badcs    lda #e_badcs          ;bad control switch
	.byte bitabs
badrange lda #e_range          ;parm range/string length
	.byte bitabs
badmode  lda #e_mode           ;wrong video mode
	.byte bitabs
badinit  lda #e_init           ;not initialized
;
;
;	process error...
;
         ldx reg_sptr          ;entry stack pointer
         txs                   ;rebalance the stack
         ldx reg_xts           ;restore .X & .Y entry values
         ldy reg_yts
         sec                   ;indicate an error &...
         bcs exit              ;return to caller
;
;===============================================================================
;
;goodexit: NON-ERROR EXIT HANDLER
;
;	
;	This function reloads the MPU registers with their entry values,
;	clears carry to indicate that no error occurred & falls through
;	to EXIT to return to the calling program
;	
;
goodexit lda #0
         pha                   ;status register exit value
         ldy reg_yts           ;restore .Y
         ldx reg_xts           ;restore .X
         lda reg_ats           ;restore .A
         plp                   ;condition status register
;
;===============================================================================
;
;exit: PROGRAM EXIT HANDLER
;
;	
;	Upon arriving here, the MPU registers will have been  conditioned
;	according to what transpired earlier.  All we're doing is putting
;	the system back the way we found it & returning parameters to the
;	calling program.
;	
;
exit     php                   ;save flags
         pha                   ;save exit code, if any
         txa                   ;save this guy as well
         pha
;
;	
;	Restore page zero
;	
;
         lda imparmx
         sta imparm
         lda imparmx+s_byte
         sta imparm+s_byte
         ldx #s_zpptrx-1
;
.0000010 lda zpptrx,x          ;ZP storage
         sta zpptra,x          ;copy to live ZP
         dex
         bpl .0000010
;
;	
;	Restore system state & exit
;	
;
         lda mmucrx
         sta mmucr             ;restore old memory map
         lda #0
         sta semaphor          ;now safe for IRQ access
         pla                   ;restore registers
         tax
         pla
         plp
         rts                   ;back to caller
;
;===============================================================================
;
;dpymgr: EXECUTE NUMBERED DISPLAY MANAGER FUNCTIONS
;
;	
;	.A = function number
;	.X = function-specific parameter
;	.Y = function-specific parameter
;	
;	.C = 0: okay
;	     1: error
;	
;
;	Initialization returns:
;	
;	.A = 16: 16K VRAM size
;	     64: 64K VRAM size
;	.X = number of screen buffers
;	.Y = 0: minimal functionality
;	     1: full functionality
;	
;
;	
;	All other functions return the entry register values if no error
;	occurs.
;
;	The function number should be ORed with $80 to wait for VRT before
;	performing the desired operation.  VRT will be ignored if it is not
;	appropriate for the desired operation.  Performance may be affected
;	by waiting for VRT.
;	
;
dpymgr   clc                   ;not the status line driver
         jsr enter             ;set up entry conditions
         cmp #funcrang         ;check function code for range
         bcs badfunc           ;bogus
;
         asl                   ;now a...
         tax                   ;function table offset
         lda funtab,x          ;get function address-1 MSB
         pha                   ;setting up launch address
         lda funtab+s_byte,x   ;get function address-1 LSB
         pha
         ldx reg_xts           ;restore
         lda funccode          ;likewise
         rts                   ;execute function
;
;===============================================================================
;
;tstsem: REPORT SEMAPHORE STATUS
;
;	
;	no entry values
;	
;	.A = entry value
;	.X = entry value
;	.Y = entry value
;	.C = 0: semaphore lowered
;	     1: semaphore raised
;	
;	Testing of the semaphore also occurs in ENTER, with a busy status
;	being returned to the calling program if another  manager call is in
;	progress.  Any calling program that wishes to know in advance if the
;	manager is busy can call this function to check status.
;	
;
tstsem   pha
         lda semaphor          ;get semaphore state
         asl                   ;condition carry
         pla
         rts
;
;===============================================================================
;
;init: INITIALIZE DISPLAY MANAGER
;
;	
;	.X = not used
;	.Y = not used
;	
;	.A = 16  16K VRAM detected
;	     64  64K VRAM detected
;	.X = available screen buffers
;	.Y = 0   minimal functionality
;	     1   full functionality
;	
;	If the display manager has already been initialized only the status
;	information will be returned.
;
;	WARNING!  Do not call INIT from within an IRQ handler unless 80CDM
;	          has already been initialized!
;	
;	
init     lda maxmem0+s_byte    ;top-of-BASIC MSB
         cmp #>_origin_
         bcc .0000010          ;below us, don't change anything
;
         lda maxmem0           ;top-of-BASIC LSB
         cmp #<_origin_
         bcc .0000010          ;below us
;
         ldx #<_origin_        ;manager load address LSB &...
         ldy #>_origin_        ;MSB
         stx maxmem0           ;move top-of-BASIC to...
         sty maxmem0+s_byte    ;protect us from big programs
;
.0000010 bit dmstatus          ;current init state
         bpl .0000020          ;not initialized
;
         jmp .0000140          ;just report status & exit
;
;	
;	Clear Working Storage
;	
;
.0000020 lda #0                ;assume 16K VRAM
         ldx #<_sstart_        ;start of cleared area LSB
         ldy #>_sstart_        ;start of cleared area MSB
         sta dmstatus          ;init state = "not initialized"
;
.0000030 stx imparm            ;set working pointer
         sty imparm+s_byte
         tay
         sta (imparm),y        ;clear
         ldy imparm+s_byte
         inx
         bne .0000040
;
         iny
;
.0000040 cpy #>_send_          ;end of cleared area MSB
         bne .0000030          ;continue
;
         cpx #<_send_          ;end of cleared area LSB
         bne .0000030          ;not done
;
;	
;	Size Video RAM
;	
;
         ldx #kramad           ;KRAM & VRAM type reg
         jsr getreg            ;get it...
         and #krdbmask         ;mask dead bits &...
         sta kramregv          ;stash it
         ora #testvrt          ;configure for test
         jsr putreg            ;tell VDC it has 64K VRAM
         ldx #<testadra        ;1st VRAM test address LSB &...
         ldy #>testadra        ;MSB
         jsr getram            ;read whatever is there &...
         sta tramvala          ;store it for later
         ldx #<testadrb        ;2nd VRAM test address LSB &...
         ldy #>testadrb        ;MSB
         jsr getram            ;read &...
         sta tramvalb          ;store it as well
         lda #testpata         ;test pattern #1
         ldx #<testadra        ;1st VRAM test address LSB &...
         ldy #>testadra        ;MSB
         jsr putram            ;write test pattern
         ldx #<testadra        ;1st VRAM test address LSB &...
         ldy #>testadra        ;MSB
         jsr getram            ;read back test pattern &...
         sta rbitpata          ;store it for later
         ldx #<testadrb        ;2nd VRAM test address LSB &...
         ldy #>testadrb        ;MSB
         jsr getram            ;read &...
         sta rbitpatb          ;store it as well
         lda #testpatb         ;test pattern #2
         ldx #<testadra        ;1st VRAM test address LSB &...
         ldy #>testadra        ;MSB
         jsr putram            ;write test pattern
         ldx #<testadra        ;1st VRAM test address LSB &...
         ldy #>testadra        ;MSB
         jsr getram            ;read back test pattern &...
         sta rbitpatc          ;store it for later
         ldx #<testadrb        ;2nd VRAM test address LSB &...
         ldy #>testadrb        ;MSB
         jsr getram            ;read &...
         sta rbitpatd          ;store it as well
         lda tramvala
         ldx #<testadra
         ldy #>testadra
         jsr putram            ;restore test location #1
         lda tramvalb
         ldx #<testadrb
         ldy #>testadrb
         jsr putram            ;restore test location #2
;
;	
;	Evaluate Test Results
;	
;
         lda rbitpata          ;test value #1 lo
         cmp rbitpatb          ;test value #1 hi
         bne .0000050          ;64K VRAM detected
;
         lda rbitpatc          ;test value #2 lo
         cmp rbitpatd          ;test value #2 hi
         beq .0000070          ;16K VRAM detected
;
.0000050 lda #sw64mask
         sta dmstatus          ;is a 64K VRAM system
         lda kramregv          ;original KRAM value
         and #testvrt          ;mask for VRAM type bit
         bne .0000060          ;4464 DRAM defined by kernel
;
         lda dmstatus
         ora #testvrt
         sta dmstatus          ;must reload fonts & clear screen
;
.0000060 lda #n_scrn64         ;user-accessible screens (64K)
         .byte bitabs
;
.0000070 lda #n_scrn16         ;user-accessible screens (16K)
         sta maxscrn           ;set user-accessible screens
         cmp #n_scrn64
         beq .0000080          ;64K system
;
         lda kramregv
         ldx #kramad
         jsr putreg            ;restore old KRAM register value
;
;	
;	Capture VDC Register Entry Values
;	
;
.0000080 ldy #s_vdcstg-1       ;number of registers to capture
;
.0000090 ldx vdcregs,y         ;register list table
         jsr getreg            ;read &...
         sta vdcregss,y        ;save
         dey
         bpl .0000090          ;next
;
;	
;	compute screen size & block copy data
;	
;
         ldx colvisx           ;visible columns on main screen
         ldy rowvisx           ;visible rows on main screen
         jsr dpmult            ;compute visible characters
         stx scrnsize          ;screen size LSB
         sty scrnsize+s_byte   ;screen size MSB
         jsr cbccit            ;compute block copy parameters
         stx bccount           ;save bytes per iteration
         sty bcitrtns          ;save iterations per copy
;
;	
;	Continue with Initialization
;	
;
         lda #bufpag16         ;16K frame buffer start page
         ldx #n_fram16         ;number of 16K frames
         bit dmstatus
         bvc .0000100          ;16K system, skip status line setup
;
;	
;	Reconfigure Display to Support Status Line
;	
;
         lda amemhix           ;default base attribute page
         sta vdcbufsa          ;attribute copy source
         lda #altarpag         ;our relocated ARAM page
         sta vm4               ;tell screen kernel about it
         sta vdcbufta          ;also target for attribute copy
         pha                   ;stash for later use
         clc
         lda scrnsize          ;screen size in chars LSB
         adc dmemlox           ;main display start LSB
         sta dslinlo           ;status line display start LSB
         lda scrnsize+s_byte   ;screen size in chars MSB
         adc dmemhix           ;main display start MSB
         sta dslinhi           ;status line display start MSB
         clc
         lda scrnsize          ;screen size in chars LSB
         adc #0                ;main attribute start LSB
         sta aslinlo           ;status line attribute start LSB
         lda scrnsize+s_byte   ;screen size in chars MSB
         adc vm4               ;main attribute start MSB
         sta aslinhi           ;status line display start MSB
         jsr cpysega           ;copy attributes to new location
         pla                   ;new attribute page
         ldx #amemhi
         jsr putreg            ;tell VDC about it
         lda #bufpag64         ;64K system start page
         ldx #n_fram64         ;number of buffer frames
;
;	
;	Set Up Screen Frame Buffer Pages
;	
;
.0000100 sta basbpag           ;set screen buffer base page
         lsr                   ;convert to...
         lsr                   ;base index
         lsr
         lsr
         sta fac1              ;base frame page index
;
.0000110 sec
         txa
         sbc #1                ;working frame index
         clc
         adc fac1              ;base frame index
	.rept 4               ;convert to page
         asl
	.endr
         sta vm3tab,x          ;save in display table slot
         clc
         adc #>s_bufseg        ;frame segment size
         sta vm4tab,x          ;save in attribute table slot
         dex
         bne .0000110          ;next frame
;
         lda vm3               ;live display start page
         sta vm3tab,x          ;save in table
         lda vm4               ;live attribute start page
         sta vm4tab,x          ;also save
         lda fgbgatx           ;fore/background color
         sta sfbtab,x          ;save in table
         lda #vssfmask
         sta ssflags,x         ;live screen always valid
         jsr cpylspdt          ;copy it to data table
         lda dmstatus          ;status word
         and #testvrt          ;modified machine?
         beq .0000130          ;no
;
;	
;	Reload Fonts if Modified with 4464 DRAM
;	
;
         ldx #dlvfonte-dlvfont-1
;
.0000120 lda dlvfont,x         ;copy font restoration code to...
         sta hwstack,x         ;low common RAM
         dex
         bpl .0000120
;
         jsr hwstack           ;reload VDC fonts & clear screen
;
;	
;	Finalize Initialization
;	
;
.0000130 lda dmstatus
         and #sw64mask         ;clean up unneeded flags
         ora #swfimask
         sta dmstatus          ;now fully initialized
;         
.0000140 bit dmstatus          ;check VRAM size
         bvc .0000150          ;16K system
;
         lda #ramlarge         ;64K VRAM
         ldy #func_all         ;all functions available
         bne .0000160
;
.0000150 lda #ramsmall         ;16K VRAM
         ldy #func_min         ;display blanker only
;
.0000160 ldx maxscrn           ;maximum user-accessible screens
         sta reg_ats           ;set exit values
         stx reg_xts
         sty reg_yts
         jmp goodexit          ;return to caller
;
;================================================================================
;
;dsline: WRITE CHARACTER STRING TO STATUS LINE
;
;	
;	.A = starting column 
;	.X = string address LSB
;	.Y = string address MSB
;	
;	.C = 0: okay
;	     1: error
;	
;	COLUMN = Zero-based column at which to start writing.  If bit 7 is set,
;	         the status line will be cleared prior to writing.
;
;	STRING = Starting address of null-terminated character string.  An err-
;	         or will occur if the sum of the string length & column exceeds
;	         the visible screen width as set in the VDC.  This error will
;	         also be returned if the null terminator is missing.
;
;	To only clear the status line call this function as follows:
;
;	         LDX #0
;	         LDY #0
;	         JSR DSLINE
;
;	.A can be anything  it isn't checked.
;
;	Access is permitted by an IRQ handler when the semaphore is clear.
;	
;
dsline   sec                   ;not a numbered function call
         jsr enter             ;do entry setup
         lda dmstatus          ;status word
         lsr                   ;check status line flag
         bcc dsline07          ;not enabled, error
;
         stx imparm            ;string pointer LSB
         sty imparm+s_byte     ;string pointer MSB
         txa
         ora imparm+s_byte     ;check for null pointer
         bne dsline01          ;writing a string
;
;	
;	Clear Status Line Only
;	
;
         sec                   ;clearing status line
         jsr clrsln
         bcc dsline06          ;done
;
;	
;	Check String Length
;	
;
dsline01 ldx funccode          ;starting column position
;
dsline02 jsr getchr            ;read the string
         bcs dsline03          ;end
;
         cpx colvisx           ;max display columns
         beq dsline08          ;out of range
;
         inx
         bne dsline02          ;keep going
;
;	
;	Write to Status Line
;	
;
dsline03 asl vrtflag           ;get clear flag
         jsr clrsln            ;prepare status line
         lda dslinlo           ;status line display VRAM LSB
         ldy dslinhi           ;status line display VRAM MSB
         adc funccode          ;adjust to starting column
         bcc dsline04
;
         iny
;
dsline04 tax
         jsr setvradr          ;set display VRAM address
         ldx reg_xts
         ldy reg_yts
         stx imparm            ;reset string pointers
         sty imparm+s_byte
;
dsline05 jsr getchr            ;get character
         bcs dsline06          ;end-of-string
;
         jsr petscode          ;convert to screen code
         jsr writvram          ;write to status line
         bcc dsline05          ;next
;
dsline06 jmp goodexit          ;done
;
dsline07 jmp badfunc           ;status line not available
;
dsline08 jmp badrange          ;string too long
;
;===============================================================================
;
;scnadj: ADJUST VERTICAL SCREEN POSITION
;
;	
;	.X = $00: reset to center
;	     $01: move display up
;	     $80: move display down
;	.Y = not used
;	
;	No error returns.
;	
;	Each call to this function will incremently adjust the display position
;	unless reset-to-center is selected.  No range checking is implemented.
;	If the user over-adjusts, the display will turn into baby barf, giving
;	him a clue that he's overdone it.
;	
;
scnadj   lda dmstatus
         and #slonmask         ;check if status line is on
         beq dsline07          ;not on, why are we here?
;
         txa
         bne scnadj01          ;make an adjustment
;
;	
;	Restore Display to Default Position
;	
;
         lda vsyposx           ;default adjustment
         bne scnadj03          ;re-center display
;
;	
;	Presumably Making an Adjustment
;	
;
scnadj01 ldx #vsypos           ;vert sync position register
         jsr getreg            ;get current setting
         ldx reg_xts           ;flag value
         cpx #vcadjswd         ;check desired operation
         beq scnadj02          ;adjust downward
;
         dex
         bne scnadj04          ;invalid switch
;
;	
;	Adjust Screen Upward
;	
;
         sec
         sbc #vctradji         ;kick it up a notch
         jmp scnadj03
;
;	
;	Adjust Screen Downward
;	
;
scnadj02 clc
         adc #vctradji         ;move down a notch
;
;	
;	Change VDC Setting
;	
;
scnadj03 ldx #vsypos           ;select register
         jsr vrwait            ;wait for VRT if requested
         jsr putreg            ;adjust vertical position
         jmp goodexit          ;done
;
scnadj04 jmp badcs             ;bad switch
;
;===============================================================================
;
;scnblk: ENABLE or DISABLE VISIBLE DISPLAY
;
;	
;	.X = 0: enable display
;	     1: disable display
;	.Y = not used
;	
;
scnblk   txa                   ;turn on display?
         beq scnblk01          ;yes
;
         lda #dpyblnkl         ;left blank value
         ldy #dpyblnkr         ;right blank value
         dex                   ;check switch
         bne scnblk03          ;bogus
;
         beq scnblk02          ;okay
;
scnblk01 lda hbposlx           ;default VDC left blank &...
         ldy hbposrx           ;right
;
scnblk02 ldx #hbposl           ;left hor pos blank reg
         jsr vrwait            ;wait for vertical retrace
         jsr putreg            ;set left side value
         tya                   ;get right value
         ldx #hbposr           ;right hor pos blank reg
         jsr putreg            ;set right side value
         jmp goodexit
;
scnblk03 jmp badcs             ;invalid switch
;
;===============================================================================
;
;scncpy: COPY SCREEN to SCREEN
;
;	
;	.X = Source Screen
;	.Y = Target Screen
;	
;
scncpy   jsr chkscn            ;check screen numbers
         txa                   ;source screen
         jsr chkssv            ;valid?
         bcs scncpy02          ;no, copy not allowed
;
         jsr cpybuf            ;copy display & attribute VRAM
         bcs scncpy03          ;huh???
;
         jsr cpysdp            ;copy ZP & shadow values
         lda screent           ;target screen
         bne scncpy01          ;not live display
;
         jsr setvcpos          ;update hardware cursor position
;
scncpy01 jmp goodexit          ;done
;
scncpy02 jmp badsrc            ;source has no data
;
scncpy03 jmp badseqt           ;target = source
;
;===============================================================================
;
;scnrdr: REDIRECT DISPLAY OUTPUT
;
;	
;	.X = not used
;	.Y = target screen
;	
;
scnrdr   cpy rdrscrn           ;same screen as last?
         beq scnrdr04          ;yes, so silently exit
;
         tya                   ;target screen
         beq scnrdr01          ;live screen always valid
;
         sec
         sbc #1                ;make target screen an index
         jsr chksnr            ;check screen number range
         bcs scnrdr05          ;out of range
;
         lda #livescrn
;
scnrdr01 sta screens           ;live will be copy source
         lda rdrscrn           ;previous redir screen
         sta screent           ;set copy destination
         bne scnrdr02          ;was not live screen
;
         jsr cpylspdt          ;transfer to live table slot
         jmp scnrdr03
;
scnrdr02 jsr cpysdp            ;transfer live to previous
;
scnrdr03 lda reg_yts           ;reload target
         sta rdrscrn           ;set new redir screen
         sta screent           ;also new copy target
         jsr scrnpage          ;get frame pages
         stx vm3               ;set display VRAM page
         sty vm4               ;set atttribute VRAM page
         jsr cvro              ;compute VRAM relative offset
         stx rowdadr           ;display address LSB
         stx rowaadr           ;attribute address LSB
         clc
         tya
         adc vm3               ;start of display RAM
         sta rowdadr+s_byte    ;absolute display address MSB
         clc
         tya
         adc vm4               ;start of attribute RAM
         sta rowaadr+s_byte    ;absolute attribute address MSB
         jsr setvcpos          ;reset VDC cursor position
         ldx rdrscrn           ;new redir screen
         lda ssflags,x         ;screen...
         ora #vsbfmask         ;buffer frame...
         sta ssflags,x         ;now valid
         txa
         beq scnrdr04          ;live screen is redir screen
;
         jsr cpysdp            ;initial ZP & cursor storage
;
scnrdr04 jmp goodexit          ;done
;
scnrdr05 jmp badtgt            ;bad target screen
;
;===============================================================================
;
;scnrts: COPY RAM to SCREEN
;
;	
;	.X = RAM-0 starting page
;	.Y = target screen
;	
;
scnrts   tya                   ;target screen
         jsr chksnr            ;check for range
         bcc scnrts02          ;in range
;
scnrts01 jmp badsrc            ;invalid source
;
scnrts02 sty screent           ;save target screen
         jsr scrnpage          ;get target buffer segment
         stx vdcbufsd          ;save display segment MSB
         sty vdcbufsa          ;save attribute segment MSB
         ldx #0                ;RAM address LSB
         ldy reg_xts           ;RAM address MSB
         stx imparm            ;set
         sty imparm+s_byte
;
;	
;	Load Screen Parameters
;	
;
         lda #r0mask
         sta mmucr             ;map out I/O
         ldy #1
         lda (imparm),y        ;number of bytes LSB
         sta zpptrc            ;set count for copy
         iny
         lda (imparm),y        ;number of bytes MSB
         sta zpptrc+s_byte
         lda #r0iomask
         sta mmucr             ;restore I/O
;
;	
;	Copy Screen Data
;	
;
         clc
         txa                   ;RAM pointer LSB
         ldy imparm+s_byte     ;RAM pointer MSB
         adc #s_byte+s_word    ;align
         bcc scnrts03
;
         iny
;
scnrts03 tax                   ;start address LSB
         lda vdcbufsd          ;display VRAM start MSB
         jsr rtov              ;copy RAM to display
         lda vdcbufsa          ;attribute VRAM start MSB
         jsr rtov              ;copy RAM to attributes
;
;	
;	Copy Kernel Display Parameters
;	
;
         lda #r0mask
         sta mmucr             ;map out I/O
         ldy screent           ;target screen
         jsr seltgt            ;set up target
         ldy #s_szptab-1       ;ZP bytes to copy -1
;
scnrts04 lda (imparm),y        ;ZP data
         sta (zpptrb),y        ;save
         dey
         bpl scnrts04
;
         clc
         lda imparm
         adc #s_szptab         ;adjust pointer
         sta imparm
         lda imparm+s_byte
         adc #0
         sta imparm+s_byte
         ldy #s_lnktab-1       ;link/tab data to copy -1
;
scnrts05 lda (imparm),y        ;link/tab data
         sta (zpptrf),y        ;save
         dey
         bpl scnrts05
;
         clc
         lda imparm
         adc #s_lnktab         ;adjust pointer
         sta imparm
         lda imparm+s_byte
         adc #0
         sta imparm+s_byte
         iny
         lda (imparm),y        ;cursor shadow
         sta (zpptrd),y        ;save
         iny
         ldx screent           ;target screen
         lda (imparm),y        ;fore/background
         sta sfbtab,x          ;update table
         lda #r0iomask
         sta mmucr             ;restore I/O
         txa         
         bne scnrts06          ;not live display
;
         lda sfbtab,x          ;fore/background value
         ldx #fgbgat           ;fore/background register
         jsr putreg            ;update display
         jsr cpylspdt          ;also update live table slots
; 
scnrts06 ldx screent           ;target screen
         lda ssflags,x         ;corresponding valid flag
         ora #vssfmask
         sta ssflags,x         ;now a valid screen
         jmp goodexit
;
;===============================================================================
;
;scnsln: ENABLE or DISABLE STATUS LINE
;
;	
;	.X = 0: status line off
;	     1: status line on
;	.Y = status line attributes...
;
;	     x000xxxx
;	     |   ||||
;	     |   ++++> background color (RGBI)
;	     +> 1 = select mixed charset
;
;	Detailed operation is as follows:
;
;	.X = 0 & SL off: no effect
;	.X = 0 & SL on:  turn off
;	.X = 1 & SL off: turn on & set attributes
;	.X = 1 & SL on:  set attributes only  
;	
;	
scnsln   bit dmstatus
         bvc scnsln06          ;function not allowed
;
         cpx #switchhi+s_byte  ;check switch
         bcs scnsln07          ;out of range
;
         txa
         beq scnsln02          ;turn off status line
;
         tya                   ;color attribute
         and #slatmask         ;should be charset & color
         bne scnsln08          ;junk
;
;	
;	Turn On or Refresh
;	
;
         tya                   ;charset & color
         ora #slrvmask         ;add in reverse video
         sta slinattr          ;now attribute
         lda dmstatus          ;status word
         lsr                   ;check status line flag bit
         bcc scnsln01          ;status line off, turn on
;
;	
;	Refresh Only
;	
;
         clc                   ;only refresh attributes
         jsr clrsln            ;refresh attributes &...
         jmp goodexit          ;exit
;
;	
;	Do Full Turn-On
;	
;
scnsln01 sec
         jsr clrsln            ;clear VRAM & set attributes
         lda #slonmask         ;"on" status word mask
         pha                   ;save
         ldy rowvisx           ;default visible rows
         iny                   ;add a row
         bne scnsln03          ;jump ahead
;
;	
;	Turn Off
;	
;
scnsln02 lda #slofmask         ;"off" status word mask
         pha                   ;save
         ldy rowvisx           ;default visible rows
;
;	
;	Adjust Display as Required
;	
;         
scnsln03 tya                   ;number of rows
         ldx #rowvis           ;select register
         jsr putreg            ;change row count
         pla                   ;get status word mask
         cmp #slonmask         ;"on" mask
         bne scnsln04
;
         ora dmstatus          ;status line is on
         bne scnsln05
;
scnsln04 and dmstatus          ;status line is off
;
scnsln05 sta dmstatus          ;condition status word &...
         jmp goodexit          ;exit
;
scnsln06 jmp badfunc           ;function not allowed
;
scnsln07 jmp badcs             ;bad switch
;
scnsln08 jmp badrange          ;bad color byte
;
;===============================================================================
;
;scnstr: COPY SCREEN TO RAM
;
;	
;	.X = source screen
;	.Y = RAM-0 starting page
;	
;	.A = RAM-0 starting page
;	.X = end address+1 LSB
;	.Y = end address+1 MSB
;	
;	Screen data is stored as follows:
;
;	Offset  Data
;	
;	 $0000   Source screen number.
;	 $0001   Number of display/attribute bytes copied LSB/MSB.
;	 $0003   Display screen code.
;	 $nnnn   Display attributes, where $nnnn is ($0001) + $0003.
;	 $pppp   Kernel parameters, where $pppp is ($0001) + $nnnn.
;	
;
scnstr   txa                   ;source screen
         jsr chksnr            ;check for range
         bcc scnstr02          ;in range
;
scnstr01 jmp badsrc            ;invalid source
;
scnstr02 jsr chkssv            ;valid data?
         bcs scnstr01          ;no, abort
;
         stx screens           ;save source screen
         lda #%10000000
         sta screent           ;target not a real screen
         txa                   ;source screen
         pha
         jsr scrnpage          ;get source buffer segment
         stx vdcbufsd          ;save display segment MSB
         sty vdcbufsa          ;save attribute segment MSB
         ldx #0                ;RAM address LSB
         ldy reg_yts           ;RAM address MSB
         stx imparm            ;set
         sty imparm+s_byte
         lda #r0mask
         sta mmucr             ;map out I/O
;
;	
;	Store Screen Parameters
;	
;
         ldy #0
         pla                   ;source screen number
         pha                   ;resave
         sta (imparm),y        ;save
         iny
         lda scrnsize          ;screen size LSB
         sta (imparm),y        ;save
         iny
         lda scrnsize+s_byte   ;screen size MSB
         sta (imparm),y        ;also save
         lda #r0iomask
         sta mmucr             ;restore I/O
;
;	
;	Copy Screen Data
;	
;
         clc
         txa                   ;RAM pointer LSB
         ldy imparm+s_byte     ;RAM pointer MSB
         adc #s_byte+s_word    ;align
         bcc scnstr03
;
         iny
;
scnstr03 tax                   ;start address LSB
         lda vdcbufsd          ;display VRAM start MSB
         jsr vtor              ;copy display to RAM
         lda vdcbufsa          ;attribute VRAM start MSB
         jsr vtor              ;copy attributes to RAM
;
;	
;	Copy Kernel Display Parameters
;	
;
         lda #r0mask
         sta mmucr             ;map out I/O
         pla                   ;source screen
         pha                   ;resave
         tax
         jsr selsrc            ;set up source
         ldy #s_szptab-1       ;ZP bytes to copy -1
;
scnstr04 lda (zpptra),y        ;ZP data
         sta (imparm),y        ;save
         dey
         bpl scnstr04
;
         clc
         lda imparm
         adc #s_szptab         ;adjust pointer
         sta imparm
         lda imparm+s_byte
         adc #0
         sta imparm+s_byte
         ldy #s_lnktab-1       ;link/tab data to copy -1
;
scnstr05 lda (zpptre),y        ;link/tab data
         sta (imparm),y        ;save
         dey
         bpl scnstr05
;
         clc
         lda imparm
         adc #s_lnktab         ;adjust pointer
         sta imparm
         lda imparm+s_byte
         adc #0
         sta imparm+s_byte
         iny
         lda (zpptrc),y        ;cursor shadow
         sta (imparm),y        ;save
         inc imparm            ;next location
         bne scnstr06
;
         inc imparm+s_byte
;
scnstr06 pla                   ;source screen
         tax
         bne scnstr07          ;not live display
;
         lda #r0iomask
         sta mmucr             ;restore I/O
         ldx #fgbgat           ;fore/background register
         jsr getreg            ;read it &...
         ldx screens
         sta sfbtab,x          ;update table
         lda #r0mask
         sta mmucr             ;map out I/O
;
scnstr07 lda sfbtab,x          ;source fore/background
         sta (imparm),y        ;save
         lda #r0iomask
         sta mmucr             ;restore I/O
;
;	
;	Finish Up
;	
;
         ldx imparm            ;final pointer
         ldy imparm+s_byte
         inx                   ;ending address +1 LSB
         bne scnstr08
;
         iny                   ;ending address +1 MSB
;
scnstr08 lda reg_yts
         sta reg_ats           ;starting RAM page
         stx reg_xts           ;end RAM address +1 LSB
         sty reg_yts           ;end RAM address +1 MSB
         jmp goodexit
;
;===============================================================================
;
;scnxch: EXCHANGE VISIBLE SCREEN with TEMPORARY SCREEN
;
;	
;	.X = not used
;	.Y = not used
;	
;	This function will abort with an invalid source screen error if screen
;	11 was not previously loaded via display output direction.
;	
;
scnxch   bit dmstatus          ;function allowed?
         bvc scnxch01          ;no, error
;        
         lda #exchscrn         ;exchange screen
         jsr chkssv            ;exchange screen valid?
         bcs scnxch02          ;no, abort
;
;	
;	Copy Live Screen to Holding Area
;	
;
         ldx #livescrn         ;source is live screen
         ldy #holdscrn         ;target is holding screen
         stx screens           ;set source
         sty screent           ;set target
         jsr cpybuf            ;copy display & attribute VRAM
         jsr cpysdp            ;copy ZP & shadow values 
;
;	
;	Copy Exchange Screen to Live Screen
;	
;
         ldx #exchscrn
         ldy #livescrn
         stx screens
         sty screent
         jsr cpybuf
         jsr cpysdp
         jsr setvcpos          ;update hardware cursor position
;
;	
;	Copy Holding Area to Exchange Screen
;	
;
         ldx #holdscrn
         ldy #exchscrn
         stx screens
         sty screent
         jsr cpybuf
         jsr cpysdp
         jmp goodexit
;
scnxch01 jmp badfunc           ;function not allowed
;
scnxch02 jmp badsrc            ;exchange has no data
;
;===============================================================================
;
;cbccit: COMPUTE BLOCK COPY COUNT & ITERATIONS
;
;	
;	.X = total bytes-to-copy LSB
;	.Y = total bytes-to-copy MSB
;	
;	.X = bytes count per iteration
;	.Y = iterations to complete copy
;	.C = 0: byte count per iteration not rounded
;	     1: byte count per iteration rounded up
;	
;	All registers used
;	
;	This subroutine computes the parameters needed to perform a VRAM block
;	copy operation.   The bytes-to-copy per iteration result is rounded up
;	as necessary.  The maximum possible bytes-to-copy should not exceed
;	2048 or undefined results will occur.
;	
;
cbccit   txa
         pha                   ;save bytes-to-copy LBS
         stx fac1              ;store bytes-to-copy LSB &...
         sty fac1+s_byte       ;MSB
         txa                   ;LSB
         ldx #s_bits           ;bits in a byte
;
;	
;	Divide Bytes-to-Copy by 256 to Get Iterations
;	
;
cbccit01 lsr fac1+s_byte
         ror fac1
         dex
         bne cbccit01
;
         lda fac1
         adc #0                ;account for carry
         sta fac2              ;copy iterations to FAC2
         lda #0
         sta fac2+s_byte       ;MSB always zero
         pla
         sta fac1              ;bytes to copy LSB
         sty fac1+s_byte       ;bytes to copy MSB
         lda #0
         sta fac1+s_word       ;clear bits 16-31 of...
         sta fac1+s_xword      ;dividend
;
;	
;	Divide Bytes-to-Copy by Iterations to Get Bytes per Iteration
;	
;
         ldx #s_bits*s_word    ;bits to process
;
cbccit02 rol fac1              ;rotate dividend
         rol fac1+s_byte
         rol fac1+s_word
         rol fac1+s_xword
         sec
         lda fac1+s_word
         sbc fac2              ;subtract divisor from dividend LSB
         tay
         lda fac1+s_xword
         sbc fac2+s_byte       ;subtract divisor from dividend MSB
         bcc cbccit03          ;discard above, not needed
;
         sty fac1+s_word       ;save partial quotient bits 16-23
         sta fac1+s_xword      ;save partial quotient bits 24-31
;
cbccit03 dex
         bne cbccit02          ;next bit
;
         rol fac1              ;rotate in last carry to...
         rol fac1+s_byte       ;finish quotient
         lda fac1+2            ;any remainder?
         beq cbccit04          ;no, rounding not required
;
         inc fac1              ;kick it up a notch
;
cbccit04 ldx fac1              ;bytes per iteration
         ldy fac2              ;total iterations
         lsr                   ;condition .Z if rounding required
         rts
;
;===============================================================================
;
;chkscn: CHECK & STORE SCREEN NUMBER PAIRS
;
;	
;	.X = source screen number
;	.Y = target screen number
;	
;	.A = entry value
;	.X = entry value
;	.Y = entry value
;	
;	This subroutine checks the validity of the screen numbers in .X & .Y &
;	stores them in SCREENS & SCREENT, respectively.  If the validity check
;	fails, an immediate abort to the appropriate error handler will occur.
;	
;
chkscn   pha
         txa
         jsr chksnr            ;check source for range
         bcs chkscn01          ;out of range
;
         sta screens           ;save new source screen
         tya
         jsr chksnr            ;check target for range
         bcs chkscn02          ;out of range
;
         sta screent           ;save new target screen
         pla
         rts                   ;back to caller
;
;	
;	Error Handler Logic
;	
;
chkscn01 lda #0                ;flag source screen invalid
	.byte bitabs
;
chkscn02 lda #1                ;flag target screen invalid
	.byte bitabs
;
chkscn03 lda #$80              ;flag target = source
         tax
         beq chkscn04          ;source out of range
;
         bpl chkscn05          ;target out of range
;
         jmp badseqt           ;target = source
;
chkscn04 jmp badsrc            ;bad source
;
chkscn05 jmp badtgt            ;bad target
;        
;===============================================================================
;
;chksnr: CHECK SCREEN NUMBER FOR RANGE
;
;	
;	.A = screen number, 0 - N
;	
;	.A = entry value
;	.X = entry value
;	.Y = entry value
;	.C = 0: screen number in range
;	     1: screen number out of range
;	
;
chksnr   cmp maxscrn           ;max available screens
         bcc .0000010          ;in range
;
         bne .0000010          ;out of range
;
         clc                   ;okay
;
.0000010 rts
;
;===============================================================================
;
;chkssv: CHECK SCREEN FOR VALIDITY
;
;	
;	.A = screen number, 0 - N
;	
;	.A = used
;	.X = screen number
;	.Y = entry value
;	.C = 0: screen valid
;	     1: screen not valid
;	
;	This subroutine determines if the screen in .A has valid data in the
;	saved screen data tables.
;	
;
chkssv   tax
         beq .0000010          ;live screen is always valid
;
         lda ssflags,x         ;get flag
         cmp #vssfmask         ;test
         bne .0000020          ;screen not valid
;
.0000010 clc
         rts
;
.0000020 sec                   ;screen is invalid
         rts
;
;================================================================================
;
;clrsln: CLEAR STATUS LINE & REFRESH ATTRIBUTES
;
;	
;	.C = 0: refresh attributes only
;	     1: clear & refresh attributes
;	
;	.A = used
;	.X = used
;	.Y = used
;	.C = 0
;	
;
clrsln   bcc .0000010          ;refresh attributes only
;
         ldx dslinlo           ;status line display VRAM LSB
         ldy dslinhi           ;status line display VRAM MSB
         jsr setvradr          ;set VRAM starting address
         lda #scblank          ;blank screen code
         jsr filsln            ;fill status line w/blanks
;
.0000010 ldx aslinlo           ;status line attribute VRAM LSB
         ldy aslinhi           ;status line attribute VRAM MSB
         jsr setvradr          ;set VRAM starting address
         lda slinattr          ;get attribute & fall through
;
;================================================================================
;
;filsln: FILL STATUS LINE DISPLAY or ATTRIBUTE VRAM
;
;	
;	.A = fill byte
;	
;	.A = entry value
;	.X = entry value
;	.Y = 0
;	.C = 0
;	
;
filsln   ldy colvisx           ;total status line columns
;
.0000010 jsr writvram          ;write
         dey
         bne .0000010          ;next
;
         clc                   ;done
         rts
;
;===============================================================================
;
;cpdaddr: COMPUTE DISPLAY VRAM ADDRESS FROM CURSOR POSITION
;
;	
;	CURXPOS = absolute column
;	CURYPOS = absolute row
;	
;	.A = entry value
;	.X = display VRAM address LSB
;	.Y = display VRAM address MSB
;	
;
cpdaddr  pha
         clc
         lda curxpos           ;absolute column
         adc rowdadr           ;row address LSB
         tax                   ;1st cell address LSB
         lda #0
         adc rowdadr+s_byte    ;row address MSB
         tay                   ;1st cell address MSB
         pla
         rts
;
;===============================================================================
;
;cpybuf: COPY FRAME BUFFER to FRAME BUFFER
;
;	
;	SCREENS = source screen number
;	SCREENT = target screen number
;	
;	.A = used
;	.X = used
;	.Y = used
;	.C = 0: copy completed
;	     1: target = source
;	
;
cpybuf   lda screens           ;source screen
         cmp screent           ;target screen
         beq .0000020          ;target = source, shouldn't happen!
;
         jsr scrnpage          ;get source buffer segment
         stx vdcbufsd          ;save source display page
         sty vdcbufsa          ;save source attribute page
         lda screent           ;target screen
         jsr scrnpage          ;get target buffer segment
         stx vdcbuftd          ;save target display page
         sty vdcbufta          ;save target attribute page
         lda screent           ;target screen
         bne .0000010          ;live not target, skip VRT test
;
         jsr vrwait            ;wait for retrace if requested
;
.0000010 jsr cpysegd           ;copy display
         jsr cpysega           ;copy attributes
         ldx screent           ;target screen
         lda ssflags,x         ;target screen flag
         ora #vsbfmask
         sta ssflags,x         ;buffer is valid
         clc
;
.0000020 rts
;
;=========================================================================
;
;cpylspdt: COPY LIVE SCREEN PARAMETERS to LIVE SCREEN DATA TABLE SLOT
;
cpylspdt jsr livissrc          ;live is source
         lda #0                ;live screen table...
         jsr seltaddr          ;is destination
         jmp cpysdpaa          ;copy
;
;=========================================================================
;
;cpydtlsp: COPY LIVE SCREEN DATA TABLE SLOT to LIVE SCREEN PARAMETERS
;
cpydtlsp lda #0                ;live screen table...
         jsr selsaddr          ;is source
         jsr livisdes          ;live is destination
         jmp cpysdpaa          ;copy
;
;===============================================================================
;
;cpysdp: COPY KERNAL DISPLAY PARAMETERS & SET FOREGROUND/BACKGROUND
;
;	
;	SCREENS = source screen number
;	SCREENT = target screen number
;	
;	.A = used
;	.X = used
;	.Y = used
;	
;
cpysdp   ldx screens           ;source
         jsr selsrc            ;select source
         ldy screent           ;target
         jsr seltgt            ;select target
;
;	----------------------------
;	CPYSDPAA is entry point used by screen exchange function.
;	----------------------------
;
cpysdpaa ldy #s_szptab-1       ;parameter bytes to copy
;
cpysdp01 lda (zpptra),y        ;window margins, etc.
         sta (zpptrb),y        ;copy
         dey
         bpl cpysdp01
;
         ldy #s_lnktab-1       ;line link & tab bitmap size -1
;
cpysdp02 lda (zpptre),y        ;line links & tabs
         sta (zpptrf),y        ;copy
         dey
         bpl cpysdp02
;
         iny
         lda (zpptrc),y        ;cursor shadow
         sta (zpptrd),y        ;copy
;
;	
;	Change Visible Fore/Background
;	
;
         ldy screent           ;target screen
         ldx screens           ;source screen
         bne cpysdp03          ;not live display
;
         ldx #fgbgat           ;fore/background register
         jsr getreg            ;read it &...
         ldx screens
         sta sfbtab,x          ;update table
; 
cpysdp03 lda sfbtab,x          ;source fore/background
         sta sfbtab,y          ;now target fore/background
         cpy #livescrn         ;check target
         bne cpysdp04          ;not live screen
;
         ldx #fgbgat           ;register
         jsr putreg            ;set fore/background
;
;	
;	Update Slot Flag
;	
;
cpysdp04 bit screent           ;destination a real screen?
         bmi cpysdp05          ;no, don't mess with flag
;
         lda ssflags,y
         ora #vbfsmask
         sta ssflags,y         ;display parms valid
;
cpysdp05 rts
;
;=========================================================================
;
;cvro: COMPUTE VRAM RELATIVE OFFSET
;
cvro     ldx colvisx           ;visible columns
         ldy curypos           ;cursor row
;
;=========================================================================
;
;dpmult: DOUBLE-PRECISION MULTIPLICATION
;
;	
;	.X = 8 bit multiplicand
;	.Y = 8 bit multiplier
;	
;	.A = used
;	.X = 16 bit product LSB
;	.Y = 16 bit product MSB
;	.C = 0
;	
;
dpmult   lda #0
         sta fac1+s_byte       ;clear product MSB
         stx fac1              ;save operands
         sty fac2
         ldx #s_bits           ;bits to multiply
;
dpmult01 asl                   ;shift partial product LSB
         rol fac1+s_byte       ;rotate partial product MSB
         asl fac2              ;shift multiplier
         bcc dpmult02          ;skip add if next bit is zero
;
         clc                   ;add partial product to...
         adc fac1              ;multiplicand
         bcc dpmult02
;
         inc fac1+s_byte       ;bump partial product MSB
;
dpmult02 dex
         bne dpmult01          ;next bit
;
         tax                   ;product LSB
         ldy fac1+s_byte       ;product MSB
         clc
         rts
;
;===============================================================================
;
;getchr: GET NEXT CHARACTER FROM NULL-TERMINATED STRING
;
;	
;	IMPARM = ZP address of null-terminated string
;	
;	.A = character or $00 if end-of-string
;	.X = entry value
;	.Y = 0
;	.C = 0: character gotten
;	     1: end-of-string
;	.N = 0: char = $01-$7F
;	     1: char = $80-$FF
;	.Z = 1: end-of-string
;	     0: character gotten
;
;	IMPARM = incremented if character is gotten
;	
;
getchr   ldy #0
         lda (imparm),y        ;get char
         beq .0000020
;
         inc imparm            ;bump pointer
         bne .0000010
;
         inc imparm+s_byte
;
.0000010 pha                   ;recondition...
         pla                   ;.N & .Z
         clc                   ;character gotten
         rts
;
.0000020 sec                   ;end-of-string detected
         rts
;
;===============================================================================
;
;livisdes: SET LIVE SCREEN AS COPY DESTINATION
;
livisdes ldx #<asvtbeg         ;ZP data
         ldy #>asvtbeg
         stx zpptrb
         sty zpptrb+s_byte
         ldx #<curmod          ;cursor shadow
         ldy #>curmod
         stx zpptrd
         sty zpptrd+s_byte
         ldx #<tabmap          ;tab & line link bitmap
         ldy #>tabmap
         stx zpptrf
         sty zpptrf+s_byte
         rts
;
;===============================================================================
;
;livissrc: SET LIVE SCREEN AS COPY SOURCE
;
;	
;	no entry values
;	
;	.A = entry value
;	.X = used
;	.Y = used
;	
;
livissrc ldx #<asvtbeg         ;ZP data
         ldy #>asvtbeg
         stx zpptra
         sty zpptra+s_byte
         ldx #<curmod          ;cursor shadow
         ldy #>curmod
         stx zpptrc
         sty zpptrc+s_byte
         ldx #<tabmap          ;line link bitmap
         ldy #>tabmap
         stx zpptre
         sty zpptre+s_byte
         rts
;
;===============================================================================
;
;petscode: CONVERT PETSCII to SCREEN CODE
;
;	
;	.A = PETSCII character
;	
;	.A = screen code
;	.X = entry value
;	.Y = entry value
;	
;
;	Conversion rules:
;
;	PETSCII Range      Input          Output      Mask
;	
;	Low Control      $00 - $1F > $00 - $1F
;	Punc & Numerals  $20 - $3F > $20 - $3F  %00111111
;	Alpha LC         $40 - $5F > $00 - $1F  %00111111
;	Alt Alpha LC     $60 - $7F > $40 - $5F  %01011111
;	High Control     $80 - $9F > $80 - $9F
;	Alpha UC         $C0 - $FE > $40 - $FE  %01111111
;	Pi               $FF > $5E
;	
;
petscode cmp #petcodcl+1       ;low control code end
         bcc .0000030          ;skip conversion
;           
         cmp #petcodl          ;regular LC range end
         bcs .0000010          ;try next
;
         and #psclmask         ;change to LC screen code
         rts                   ;done
;
.0000010 cmp #petcodm          ;alternate LC range
         bcs .0000020          ;UC range or pi
;
         and #pscmmask
         rts
;
.0000020 cmp #petcodch+1       ;high control code end
         bcc .0000030          ;no conversion
;
         and #pschmask
         cmp #petcodp          ;is it pi?
         bne .0000030          ;no, so return it unchanged
;
         lda #petcodps         ;pi screen code
;
.0000030 rts
;
;===============================================================================
;
;rtov: COPY SYSTEM RAM to VRAM
;
;	
;	.A = VRAM segment page MSB
;	.X = starting system RAM address LSB
;	.Y = starting system RAM address MSB
;	
;	.A = used
;	.X = used
;	.Y = used
;	IMPARM = end of system RAM address +1
;	
;	The number of bytes to copy is determined by the screen size stored at
;	ZPPTRC (16 bits).  An incorrect screen size value will probably cause
;	display corruption.
;	
;
rtov     stx imparm            ;set system RAM address LSB
         sty imparm+s_byte     ;set system RAM address MSB
         ldx #0                ;VRAM page start LSB
         tay                   ;VRAM page start MSB
         jsr setvradr          ;set address
         ldx #0                ;starting count LSB
         ldy #0                ;starting count MSB
         sei                   ;interrupts off for performance
;
rtov01   stx zpptrd            ;save counter
         sty zpptrd+s_byte
         ldx #r0mask
         stx mmucr             ;map out I/O
         ldy #0
         lda (imparm),y        ;fetch from RAM
         ldx #r0iomask
         stx mmucr             ;restore config
         jsr writvram          ;store in VRAM
         inc imparm            ;next location
         bne rtov02
;
         inc imparm+s_byte
;
rtov02   ldx zpptrd            ;count LSB
         ldy zpptrd+s_byte     ;count MSB
         inx                   ;bump
         bne rtov03
;
         iny
;
rtov03   cpy zpptrc+s_byte     ;screen size MSB
         bcc rtov01            ;not done
;
         cpx zpptrc            ;screen size LSB
         bcc rtov01            ;not done
;
         cli                   ;interrupts on
         ldx imparm
         ldy imparm+s_byte
         rts
;
;===============================================================================
;
;scrnpage: CONVERT SCREEN NUMBER INTO BUFFER SEGMENT PAGES
;
;	
;	.A = screen number
;	
;	.A = screen number
;	.X = display buffer segment page
;	.Y = attribute buffer segment page
;	
;
scrnpage pha                   ;protect
         tax                   ;table index
         lda vm3tab,x          ;get display page
         ldy vm4tab,x          ;get attribute page
         tax                   ;display page
         pla                   ;recover
         rts
;
;===============================================================================
;
;selsrc: SELECT SOURCE FOR DISPLAY PARAMETERS COPY
;
;	.X = source screen number, 0 - N
;	
;	ZPPTRA = ZP source start
;	ZPPTRC = shadow source start
;	
;	all registers preserved
;
selsrc   pha                   ;preserve registers
         tya
         pha
         txa
         pha
         bne .0000010          ;not live screen
;
         jsr livissrc          ;live screen is source
         jmp .0000020
;
.0000010 sec
         sbc #1
         jsr selsaddr
;
.0000020 pla
         tax
         pla
         tay
         pla
         rts
;
;===============================================================================
;
;selsaddr: SET COPY SOURCE TABLE ADDRESSES
;
selsaddr asl                   ;index
         tax
         lda szpaddr,x         ;ZP lookup table
         sta zpptra
         lda szpaddr+s_byte,x
         sta zpptra+s_byte
         lda scsaddr,x         ;shadow lookup table
         sta zpptrc
         lda scsaddr+s_byte,x
         sta zpptrc+s_byte
         lda slladdr,x         ;line link lookup table
         sta zpptre
         lda slladdr+s_byte,x
         sta zpptre+s_byte
         rts
;
;===============================================================================
;
;seltaddr: SET COPY TARGET TABLE ADDRESSES
;
seltaddr asl                   ;index
         tax
         lda szpaddr,x         ;ZP lookup table
         sta zpptrb
         lda szpaddr+s_byte,x
         sta zpptrb+s_byte
         lda scsaddr,x         ;shadow lookup table
         sta zpptrd
         lda scsaddr+s_byte,x
         sta zpptrd+s_byte
         lda slladdr,x         ;line link lookup table
         sta zpptrf
         lda slladdr+s_byte,x
         sta zpptrf+s_byte
         rts
;
;===============================================================================
;
;seltgt: SELECT TARGET FOR DISPLAY PARAMETERS COPY
;
;	
;	.Y = target screen number, 0 - N
;	
;	ZPPTRB = ZP target start
;	ZPPTRD = shadow target start
;	
;	All registers preserved
;	
;
seltgt   pha                   ;preserve registers
         txa
         pha
         tya
         pha
         bne .0000010          ;not live screen
;
         jsr livisdes          ;live is destination
         jmp .0000020
;
.0000010 sec
         sbc #1
         jsr seltaddr          ;set target addresses
;
.0000020 pla
         tay
         pla
         tax
         pla
         rts
;
;===============================================================================
;
;vtor: COPY VRAM to SYSTEM RAM
;
;	
;	.A = VRAM segment page MSB
;	.X = starting system RAM address LSB
;	.Y = starting system RAM address MSB
;	
;	.A = used
;	.X = end system RAM address+1 LSB
;	.Y = end system RAM address+1 MSB
;	
;	The number of bytes to copy is determined by the detected screen size
;	established during initialization.
;	
;
vtor     stx imparm            ;set system RAM address LSB
         sty imparm+s_byte     ;set system RAM address MSB
         ldx #0                ;VRAM page start LSB
         tay                   ;VRAM page start MSB
         jsr setvradr          ;set address
         ldx #0                ;starting count LSB
         ldy #0                ;starting count MSB
         sei                   ;interrupts off for performance
;
vtor01   stx zpptrd            ;save count
         sty zpptrd+s_byte
         jsr readvram          ;get from VRAM
         ldx #r0mask
         stx mmucr             ;map out I/O
         ldy #0
         sta (imparm),y        ;store in RAM
         ldx #r0iomask
         stx mmucr             ;restore config
         inc imparm            ;next location
         bne vtor02
;
         inc imparm+s_byte
;
vtor02   ldx zpptrd            ;count LSB
         ldy zpptrd+s_byte     ;count MSB
         inx                   ;bump
         bne vtor03
;
         iny
;
vtor03   cpy scrnsize+s_byte   ;screen size MSB
         bne vtor01            ;not done
;
         cpx scrnsize          ;screen size LSB
         bne vtor01            ;not done
;
         cli                   ;interrupts on
         ldx imparm            ;end +1 LSB
         ldy imparm+s_byte     ;end +1 MSB
         rts
;
;===============================================================================
;
;cpysega: PERFORM ATTRIBUTE SEGMENT BUFFER COPY
;
cpysega  ldx #0                ;target attribute page LSB
         ldy vdcbufta          ;target attribute page MSB
         jsr setvradr          ;set target address
         lda vdcbufsa          ;attribute source page MSB
         jmp cpsegcom
;
;===============================================================================
;
;cpysegd: PERFORM DISPLAY SEGMENT BUFFER COPY
;
cpysegd  ldx #0                ;target display page LSB
         ldy vdcbuftd          ;target display page MSB
         jsr setvradr          ;set target address
         lda vdcbufsd          ;display source page MSB
;
;===============================================================================
;
;cpsegcom: SEGMENT BUFFER COPY COMMON CODE
;
cpsegcom ldx #bsrchi           ;copy source address register MSB
         jsr putreg            ;set source buffer address MSB
;
;===============================================================================
;
;cpyblk: PERFORM FRAME BUFFER BLOCK COPY
;
cpyblk   lda #0                ;source page LSB (even page boundary)
         ldx #bsrclo           ;copy source address register LSB
         jsr putreg            ;set source buffer address LSB
         ldx #vsscr            ;vertical smooth scroll & control
         jsr getreg            ;read current
         ora #vrbcmask         ;copying, not filling
         jsr putreg            ;tell VDC about it
         lda bccount           ;bytes per copy iteration
         ldx #blkbyt           ;VDC block copy register
         ldy bcitrtns          ;required iterations
;
.0000010 jsr putreg            ;copy away
         dey
         bne .0000010          ;next
;
         rts
;
;===============================================================================
;
;dlvfont: VDC FONT DATA RESTORATION
;
;	
;	This code is copied to the bottom of the stack & executed  to  rebuild
;	the VDC font data in the event we have to reconfigure the VDC to  rec-
;	ognize all available VRAM.  The screen is cleared following  the  font
;	download.
;	
;
dlvfont  lda #romhmask
         sta mmucr             ;RAM-0 & high ROM
         jsr dlchr             ;copy char ROM to VRAM
         jsr clscrn            ;clear the screen
         lda #r0iomask
         sta mmucr             ;restore default map
         rts
dlvfonte =*                    ;place marker
;
;===============================================================================
;
;getram: GET BYTE FROM VRAM AT SPECIFIED ADDRESS
;
;	
;	.X = VRAM address LSB
;	.Y = VRAM address MSB
;	
;	.A = returned byte
;	.X = used
;	.Y = used
;	
;
getram   jsr setvradr          ;set VRAM address &...
;
;===============================================================================
;
;readvram: READ FROM VRAM
;
readvram ldx #vramrw           ;VRAM access register
;
;===============================================================================
;
;getreg: READ FROM VDC REGISTER
;
;	
;	.A = not used
;	.X = VDC register number
;	.Y = not used
;	
;	.A = register content
;	.X = entry value
;	.Y = entry value
;	
;
getreg   stx vdcadr            ;set register
;
.0000010 bit vdcadr            ;wait until...
         bpl .0000010          ;ready
;
         lda vdcdat            ;get register contents
         clc                   ;done
         rts
;
;===============================================================================
;
;putram: WRITE BYTE TO VRAM AT SPECIFIED ADDRESS
;
;	
;	.A = byte to write
;	.X = VRAM address LSB
;	.Y = VRAM address MSB
;	
;	.A = entry value
;	.X = used
;	.Y = used
;	
;
putram   jsr setvradr          ;set VRAM address &...
;
;===============================================================================
;
;writvram: WRITE TO VRAM
;
writvram ldx #vramrw           ;VRAM access register
;
;===============================================================================
;
;putreg: WRITE TO VDC REGISTER
;
;	
;	.A = content to write
;	.X = register number
;	.Y = not used
;	
;	.A = entry value
;	.X = entry value
;	.Y = entry value
;	
;
putreg   stx vdcadr            ;set register
;
.0000010 bit vdcadr            ;wait until...
         bpl .0000010          ;ready
;
         sta vdcdat            ;put register contents
         clc                   ;done
         rts
;
;===============================================================================
;
;setvcpos: SET VDC CURSOR ADDRESS
;
;	
;	CURXPOS = absolute cursor column
;	ROWDADR = cursor row VRAM address
;	
;	.A = used
;	.X = used
;	.Y = used
;	
;	This subroutine falls through to set the VRAM location to the
;	display address that is equivalent to the cursor address.
;	
;
setvcpos clc
         lda rowdadr           ;absolute row address LSB
         adc curxpos           ;absolute cursor column
         pha                   ;absolute cursor address LSB
         lda rowdadr+s_byte    ;absolute row address MSB
         adc #0                ;handle carry
         tay
         ldx #cposhi           ;cursor address MSB register 
         jsr putreg            ;set cursor address MSB
         pla                   ;address LSB
         ldx #cposlo           ;cursor address LSB register 
         jsr putreg            ;set cursor address LSB
         tax                   ;fall thru...
;
;===============================================================================
;
;setvradr: SET UP VRAM ACCESS ADDRESS
;
;	
;	.A = not used
;	.X = VRAM address LSB
;	.Y = VRAM address MSB
;	
;	.A = entry value
;	.X = entry value
;	.Y = entry value
;	
;
setvradr pha                   ;preserve registers
         txa
         pha
         tya                   ;VRAM address MSB
         ldx #vramhi           ;select VRAM address register MSB
         jsr putreg            ;set address MSB
         pla                   ;VRAM address LSB
         ldx #vramlo           ;select VRAM address register LSB
         jsr putreg            ;set address LSB
         tax                   ;restore
         pla                   ;likewise
         rts
;
;===============================================================================
;
;vrwait: WAIT FOR VERTICAL RETRACE BLANK
;
vrwait   bit vrtflag           ;check if we should wait
         bpl .0000020          ;no, just exit
;
         pha
         lda #vrtmask          ;retrace active mask
;
.0000010 bit vdcadr            ;watch status &...
         beq .0000010          ;wait til VRT bit is set
;
         pla
;
.0000020 rts                   ;done
;
;===============================================================================
;
;STATIC DATA
;
copyrit  .byte "C-128 80 COLUMN DISPLAY MANAGER V1.1",a_cr
         .byte "(C)1987 BY BCS TECHNOLOGY LIMITED",a_cr
         .byte "ALL RIGHTS RESERVED.",a_cr,a_nul
;
dmstatus .byte 0               ;display manager status word...
;	
;	xx00000x
;	||     |
;	||     +> 1 = status line enabled
;	|+> 1 = 64K VRAM available
;	+> 1 = fully initialized
;
semaphor .byte semamask        ;display manager access semaphore...
;
;	x0000000
;	|
;	+> 1 = display manager busy
;
;	
;	Numbered Function Dispatch Table
;	
;
funtab   .dbyte init-s_byte    ;initialize display manager
         .dbyte scncpy-s_byte  ;copy screen
         .dbyte scnxch-s_byte  ;exchange screens
         .dbyte scnrdr-s_byte  ;redirect output
         .dbyte scnsln-s_byte  ;enable/disable status line
         .dbyte scnblk-s_byte  ;blank screen
         .dbyte scnadj-s_byte  ;adjust display position
         .dbyte scnstr-s_byte  ;copy screen to system RAM
         .dbyte badfunc-s_byte ;unimplemented function
         .dbyte scnrts-s_byte  ;copy system RAM to screen
funtabe
;
;	
;	Captured VDC Registers Lookup
;	
;
vdcregs  .byte colvis          ;visible columns
         .byte rowtot          ;raster rows
         .byte rowvis          ;visible rows
         .byte vsypos          ;vertical sync position
         .byte dmemhi          ;display VRAM base MSB
         .byte dmemlo          ;display VRAM base LSB
         .byte amemhi          ;attribute VRAM base MSB
         .byte amemlo          ;attribute VRAM base LSB
         .byte fgbgat          ;foreground/background colors
         .byte hbposl          ;left horizontal blank position
         .byte hbposr          ;right horizontal blank position
vdcregse
;
;	
;	Saved Screen Cursor Shadow Data Lookup
;	
;
scsaddr  .word svctab+{s_byte*0}
         .word svctab+{s_byte*1}
         .word svctab+{s_byte*2}
         .word svctab+{s_byte*3}
         .word svctab+{s_byte*4}
         .word svctab+{s_byte*5}
         .word svctab+{s_byte*6}
         .word svctab+{s_byte*7}
         .word svctab+{s_byte*8}
         .word svctab+{s_byte*9}
         .word svctab+{s_byte*10}
;
;	
;	Saved Screen Line Link & Tab Stop Data Lookup
;	
;
slladdr  .word svltab+{s_tabsvl*0}
         .word svltab+{s_tabsvl*1}
         .word svltab+{s_tabsvl*2}
         .word svltab+{s_tabsvl*3}
         .word svltab+{s_tabsvl*4}
         .word svltab+{s_tabsvl*5}
         .word svltab+{s_tabsvl*6}
         .word svltab+{s_tabsvl*7}
         .word svltab+{s_tabsvl*8}
         .word svltab+{s_tabsvl*9}
         .word svltab+{s_tabsvl*10}
;
;	
;	Saved Screen ZP Data Lookup
;	
;
szpaddr  .word szptab+{s_szptab*0}
         .word szptab+{s_szptab*1}
         .word szptab+{s_szptab*2}
         .word szptab+{s_szptab*3}
         .word szptab+{s_szptab*4}
         .word szptab+{s_szptab*5}
         .word szptab+{s_szptab*6}
         .word szptab+{s_szptab*7}
         .word szptab+{s_szptab*8}
         .word szptab+{s_szptab*9}
         .word szptab+{s_szptab*10}
;
;	
;	Display Manager Presence Signature
;	
;
signatr  .byte "80CDM...BCS2008",a_nul
;
;===============================================================================
;
;DYNAMIC DATA TABLES
;
_sstart_ =*                    ;start of dynamic storage
;
ssflags  =_sstart_             ;saved screen status flags...
;	
;	xx000000
;	||
;	|+> 1 = screen frame valid
;	+> 1 = data table entry valid
;
         *=ssflags+s_tabssf
sfbtab   =*                    ;save foreground/background colors
         *=sfbtab+s_sfbtab
svctab   =*                    ;saved screen cursors
         *=svctab+s_tabsvc
svltab   =*                    ;saved screen line links & tabs
         *=svltab+s_tabsvl
szptab   =*                    ;save screen ZP parameters
         *=szptab+s_tabszp
vm3tab   =*                    ;display VRAM frame pages
         *=vm3tab+n_fram64+s_byte
vm4tab   =*                    ;attribute VRAM frame pages
         *=vm4tab+n_fram64+s_byte
;
;===============================================================================
;
;WORKING STORAGE
;
;	----------
;	VDC Register Storage
;	----------
;
vdcregss =*                    ;start of VDC register storage...
colvisx  =vdcregss             ;visible columns
rowtotx  =colvisx+s_byte       ;raster rows
rowvisx  =rowtotx+s_byte       ;visible rows
vsyposx  =rowvisx+s_byte       ;vertical sync position
dmemhix  =vsyposx+s_byte       ;display base (MSB)
dmemlox  =dmemhix+s_byte       ;display base (LSB)
amemhix  =dmemlox+s_byte       ;attribute base (MSB)
amemlox  =amemhix+s_byte       ;attribute base (LSB)
fgbgatx  =amemlox+s_byte       ;foreground/background color
hbposlx  =fgbgatx+s_byte       ;left horizontal blank position
hbposrx  =hbposlx+s_byte       ;right horizontal blank position
         *=hbposrx+s_byte          
;
;	----------
;	VRAM Sizing Workspace
;	----------
;
kramregv =vdcregss             ;place to stash current KRAM value
rbitpata =kramregv+s_byte      ;returned bit pattern from 1st test
rbitpatb =rbitpata+s_byte      ;returned bit pattern from 2nd test
rbitpatc =rbitpatb+s_byte      ;returned bit pattern from 3rd test
rbitpatd =rbitpatc+s_byte      ;returned bit pattern from 4th test
tramvala =rbitpatd+s_byte      ;byte read from VRAM test location A
tramvalb =tramvala+s_byte      ;byte read from VRAM test location B
;
;	--------------
;	Status Line Driver Parameters
;	--------------
;
aslinlo  =*                    ;attribute VRAM base (LSB)
aslinhi  =aslinlo+s_byte       ;attribute VRAM base (MSB)
dslinlo  =aslinhi+s_byte       ;display VRAM base (LSB)
dslinhi  =dslinlo+s_byte       ;display VRAM base (MSB)
slinattr =dslinhi+s_byte       ;display attribute
         *=slinattr+s_byte
;
;	----------
;	Screen Copy Workspace
;	----------
;
bccount  =*                    ;bytes per copy interation
bcitrtns =bccount+s_byte       ;computed block copy iterations
screens  =bcitrtns+s_byte      ;copy source screen number
screent  =screens+s_byte       ;copy target screen number
vdcbufsa =screent+s_byte       ;source attribute frame page
vdcbufsd =vdcbufsa+s_byte      ;source display frame page
vdcbufta =vdcbufsd+s_byte      ;target attribute frame page
vdcbuftd =vdcbufta+s_byte      ;target display frame page
         *=vdcbuftd+s_byte
;
;	-------------------
;	Miscellaneous Screen Management Values
;	-------------------
;
basbpag  =*                    ;base screen buffer page
maxscrn  =basbpag+s_byte       ;maximum saved screens allowed
rdrscrn  =maxscrn+s_byte       ;display redirection screen
scrnsize =rdrscrn+s_byte       ;screen size in characters
vrtflag  =scrnsize+s_word      ;$80 = wait for vertical retrace
         *=vrtflag+s_byte
;
_send_   =*                    ;end of storage area cleared by init
;
;	---------------
;	MPU Register Entry/Exit Storage
;	---------------
;
reg_ats  =*                    ;.A
reg_xts  =reg_ats+s_byte       ;.X
reg_yts  =reg_xts+s_byte       ;.Y
reg_sptr =reg_yts+s_byte       ;SP
         *=reg_sptr+s_byte
;
;	--------------
;	ZP Pointer Protective Storage
;	--------------
;
imparmx  =*
zpptrx   =imparmx+s_word
         *=zpptrx+s_zpptrx
;
;	----------
;	Miscellaneous Storage
;	----------
;
funccode =*                    ;function code
mmucrx   =funccode+s_byte      ;MMU config entry value
mmurcrx  =mmucrx+s_byte        ;common RAM config entry value
         *=mmurcrx+s_byte
;
;===============================================================================
	.end